"""Functions for interacting with OPUS files.

Implements: `OpusFile`, `OpusHTTPInterface`.

Read https://tccon-wiki.caltech.edu/Main/I2SAndOPUSHeaders for more information
about the file parameters. This requires you to install this utils library with
the optional `opus` dependency:

```bash
pip install "tum_esm_utils[opus]"
# or
pdm add "tum_esm_utils[opus]"
```

Credits to Friedrich Klappenbach (friedrich.klappenbach@tum.de) for decoding the OPUS file
format."""

from . import file_interface as file_interface
from .file_interface import OpusFile as OpusFile

from . import http_interface as http_interface
from .http_interface import OpusHTTPInterface as OpusHTTPInterface



"""Mathematical functions.

Implements: `distance_between_angles`"""


def distance_between_angles(angle_1: float, angle_2: float) -> float:
    """Calculate the directional distance (in degrees) between two angles."""
    if angle_1 > angle_2:
        return min(angle_1 - angle_2, 360 + angle_2 - angle_1)
    else:
        return min(angle_2 - angle_1, 360 + angle_1 - angle_2)


def divides_evenly(dividend: float, divisor: float, precision: int = 6) -> bool:
    """Check if divisor divides dividend evenly.

    Normally this shoudld be done by `dividend % divisor == 0`, but this
    can lead to floating point errors, i.e. `1 % 0.1 == 0.09999999999999998`.
    Using `math.fmod` also does not seem to work correctly with floats."""

    multiplier = dividend / divisor
    return round(multiplier, precision) == round(multiplier)



"""Functions for interacting with GitHub and GitLab.

Implements: `request_github_file`, `request_gitlab_file`"""

from typing import Optional
import json
import os
import tum_esm_utils.shell


def request_github_file(
    repository: str,
    filepath: str,
    access_token: Optional[str] = None,
    branch_name: str = "main",
    timeout: int = 10,
) -> str:
    """Sends a request and returns the content of the response, as a string.
    Raises an HTTPError if the response status code is not 200.

    Args:
        repository:    In the format "owner/repo".
        filepath:      The path to the file in the repository.
        access_token:  The GitHub access token. Only required if the repo is private.
        branch_name:   The branch name.
        timeout:       The request timeout in seconds.

    Returns:
        The content of the file as a string.
    """

    import requests

    headers = {
        "Accept": "application/text",
    }
    if access_token is not None:
        headers["Authorization"] = f"token {access_token}"

    response = requests.get(
        f"https://raw.githubusercontent.com/{repository}/{branch_name}/{filepath}",
        headers=headers,
        timeout=timeout,
        allow_redirects=True,
    )
    response.raise_for_status()
    return response.text


def download_github_release_asset(
    repository: str,
    asset_name: str,
    dst_dir: str,
    final_name: Optional[str] = None,
    access_token: Optional[str] = None,
    force: bool = False,
) -> None:
    """Downloads a specific asset from the latest release of a GitHub repository.

    Not supported on windows!

    Args:
        repository:    In the format "owner/repo".
        asset_name:    The name of the asset to download.
        dst_dir:       The directory where the asset will be saved.
        final_name:    Optional final name for the downloaded asset. If None, uses `asset_name`.
        access_token:  The GitHub access token. Only required if the repo is private.
        force:         If True, forces the download even if the file already exists.
    """

    if os.name != "posix":
        raise OSError("download_github_release_asset is only supported on Unix systems")

    if final_name is None:
        final_name = asset_name
    if os.path.exists(os.path.join(dst_dir, final_name)) and not force:
        return

    try:
        releases = json.loads(
            tum_esm_utils.shell.run_shell_command(
                (
                    f"curl -L "
                    + f'-H "Accept: application/vnd.github+json" '
                    + (f'-H "Authorization: Bearer {access_token}" ' if access_token else "")
                    + ' -H "X-GitHub-Api-Version: 2022-11-28" '
                    + f"https://api.github.com/repos/{repository}/releases"
                )
            )
        )
    except:
        raise RuntimeError(
            f"Repository '{repository}' not found or access token does not have the required permissions."
        )

    if not isinstance(releases, list):
        raise RuntimeError(
            f"Failed to fetch releases from repository '{repository}'. Response: {releases}"
        )
    for release in releases: # pyright: ignore[reportUnknownVariableType]
        if not isinstance(release, dict):
            raise RuntimeError(
                f"Invalid release format in repository '{repository}'. Release: {releases}"
            )

    url: Optional[str] = None
    for o in sorted(releases, key=lambda x: x["published_at"], reverse=True): # type: ignore
        for asset in o["assets"]: # pyright: ignore[reportUnknownVariableType]
            if asset["name"] == asset_name:
                url = asset["url"] # pyright: ignore[reportUnknownVariableType]
                break
    if not isinstance(url, str):
        raise RuntimeError(
            f"Asset '{asset_name}' not found in any release of repository '{repository}'."
        )

    tum_esm_utils.shell.run_shell_command(
        f"curl -L "
        + (f'-H "Authorization: Bearer {access_token}" ' if access_token else "")
        + f'-H "Accept: application/octet-stream" '
        + f"-o {asset_name} "
        + url,
        working_directory=dst_dir,
    )
    if final_name != asset_name:
        final_path = os.path.join(dst_dir, final_name)
        if os.path.exists(final_path):
            os.remove(final_path)
        os.rename(os.path.join(dst_dir, asset_name), final_path)


def request_gitlab_file(
    repository: str,
    filepath: str,
    access_token: Optional[str] = None,
    branch_name: str = "main",
    hostname: str = "gitlab.com",
    timeout: int = 10,
) -> str:
    """Sends a request and returns the content of the response, as a string.
    Raises an HTTPError if the response status code is not 200.

    Args:
        repository:    In the format "owner/repo".
        filepath:      The path to the file in the repository.
        access_token:  The GitLab access token. Only required if the repo is private.
        branch_name:   The branch name.
        hostname:      The GitLab hostname.
        timeout:       The request timeout in seconds.

    Returns:
        The content of the file as a string.
    """

    import requests

    auth_param: str = ""
    if access_token is not None:
        auth_param = f"?private_token={access_token}"
    response = requests.get(
        f"https://{hostname}/{repository}/-/raw/{branch_name}/{filepath}{auth_param}",
        headers={
            "Accept": "application/text",
        },
        timeout=timeout,
    )
    response.raise_for_status()
    return response.text



"""Dataframe-related utility functions.

Implements: `fill_df_time_gaps_with_nans`

This requires you to install this utils library with the optional `polars` dependency:

```bash
pip install "tum_esm_utils[polars]"
# or
pdm add "tum_esm_utils[polars]"
```
"""

import datetime
import polars as pl


def fill_df_time_gaps_with_nans(
    df: pl.DataFrame,
    time_col: str,
    max_gap_seconds: int,
) -> pl.DataFrame:
    """Fill time gaps in a dataframe with NaN rows. This is very useful for plotting dataframes where time gaps should be visible.

    Args:
        df: The input dataframe.
        time_col: The name of the time column.
        max_gap_seconds: The maximum gap in seconds to fill with NaN rows."""

    assert max_gap_seconds > 1, "max_gap_seconds must be greater than 1"
    gaps_in_df = df.select(
        time_col,
        pl.col(time_col).diff().dt.total_seconds().alias(f"{time_col}_diff_seconds"),
    ).filter(
        pl.col(f"{time_col}_diff_seconds").gt(max_gap_seconds),
    )["utc"] - datetime.timedelta(seconds=1)
    gap_df = pl.DataFrame(
        {
            time_col: gaps_in_df,
            **{c: [None] * len(gaps_in_df) for c in df.columns if c != time_col},
        }
    )
    df_with_gaps = pl.concat([df, gap_df], how="vertical").sort(time_col)
    return df_with_gaps



"""Functions used for timing or time calculations.

Implements: `date_range`, `ensure_section_duration`, `set_alarm`,
`clear_alarm`, `wait_for_condition`, `ExponentialBackoff`"""

import os
from typing import Any, Callable, Generator, Literal, Optional
import contextlib
import datetime
import re
import signal
import time
import math
import pytz


def date_range(
    from_date: datetime.date,
    to_date: datetime.date,
) -> list[datetime.date]:
    """Returns a list of dates between from_date and to_date (inclusive)."""
    delta = to_date - from_date
    assert delta.days >= 0, "from_date must be before to_date"
    return [from_date + datetime.timedelta(days=i) for i in range(delta.days + 1)]


def time_range(
    from_time: datetime.time,
    to_time: datetime.time,
    time_step: datetime.timedelta,
) -> list[datetime.time]:
    """Returns a list of times between from_time and to_time (inclusive)."""
    # timedeltas only possible on dates
    assert from_time <= to_time, "from_time must be less equal to_time"
    _d = datetime.date(2000, 1, 1)
    _datetimes = [datetime.datetime.combine(_d, from_time)]
    while True:
        next_dt = _datetimes[-1] + time_step
        if next_dt.date() > _d:
            break
        if next_dt.time() > to_time:
            break
        _datetimes.append(next_dt)
    return [d.time() for d in _datetimes]


@contextlib.contextmanager
def ensure_section_duration(duration: float) -> Generator[None, None, None]:
    """Make sure that the duration of the section is at least the given duration.

    Usage example - do one measurement every 6 seconds:

    ```python
    with ensure_section_duration(6):
        do_measurement()
    ```
    """

    start_time = time.time()
    yield
    remaining_loop_time = start_time + duration - time.time()
    if remaining_loop_time > 0:
        time.sleep(remaining_loop_time)


def set_alarm(timeout: int, label: str) -> None:
    """Set an alarm that will raise a `TimeoutError` after
    `timeout` seconds. The message will be formatted as
    `{label} took too long (timed out after {timeout} seconds)`."""

    if os.name != "posix":
        raise OSError("set_alarm is only supported on Unix systems")

    def alarm_handler(*args: Any) -> None:
        raise TimeoutError(f"{label} took too long (timed out after {timeout} seconds)")

    signal.signal(signal.SIGALRM, alarm_handler)  # type: ignore
    signal.alarm(timeout)  # type: ignore


def clear_alarm() -> None:
    """Clear the alarm set by `set_alarm`."""

    if os.name != "posix":
        raise OSError("clear_alarm is only supported on Unix systems")

    signal.alarm(0)  # type: ignore


def parse_timezone_string(
    timezone_string: str,
    dt: Optional[datetime.datetime] = None,
) -> float:
    """Parse a timezone string and return the offset in hours.

    Why does this function exist? The `strptime` function cannot parse strings
    other than "±HHMM". This function can also parse strings in the format "±H"
    ("+2", "-3", "+5.5"), and "±HH:MM".

    Examples:

    ```python
    parse_timezone_string("GMT")        # returns 0
    parse_timezone_string("GMT+2")      # returns 2
    parse_timezone_string("UTC+2.0")    # returns 2
    parse_timezone_string("UTC-02:00")  # returns -2
    ```

    You are required to pass a datetime object in case the utc offset for the
    passed timezone is not constant - e.g. for "Europe/Berlin"."""

    offset: float = 0
    zone_string: str

    # parse the offset string
    number_of_offset_signs = timezone_string.count("+") + timezone_string.count("-")
    if number_of_offset_signs == 0:
        zone_string = timezone_string
    elif number_of_offset_signs == 1:
        s2 = timezone_string.split("+") if "+" in timezone_string else timezone_string.split("-")
        assert len(s2) == 2
        zone_string = s2[0]
        offset_string = s2[1]

        # parse the offset string
        if re.match(r"^\d{1,2}(\.\d{1})?$", offset_string):
            offset = float(offset_string)
        elif re.match(r"^\d{2}:\d{2}$", offset_string):
            offset = float(offset_string.split(":")[0]) + float(offset_string.split(":")[1]) / 60
        elif re.match(r"^\d{4}$", offset_string):
            offset = float(offset_string[:2]) + float(offset_string[2:]) / 60
        else:
            raise ValueError(f'Invalid offset string: "{offset_string}"')
        if "-" in timezone_string:
            offset = -offset
    else:
        raise ValueError(f'Invalid timezone string: "{timezone_string}"')

    # parse the time zone string
    try:
        tz = pytz.timezone(zone_string)
    except pytz.UnknownTimeZoneError:
        raise ValueError(
            f'Unknown time zone: "{zone_string}", the available time zones are: {pytz.all_timezones}'
        )
    td = tz.utcoffset(dt)
    assert td is not None, (
        f'Zone "{zone_string}" requires a datetime object to calculate the offset.'
    )
    offset += td.total_seconds() / 3600
    return offset


def wait_for_condition(
    is_successful: Callable[[], bool],
    timeout_message: str,
    timeout_seconds: float = 5,
    check_interval_seconds: float = 0.25,
) -> None:
    """Wait for the given condition to be true, or raise a TimeoutError
    if the condition is not met within the given timeout. The condition
    is passed as a function that will be called periodically.

    Args:
        is_successful:             A function that returns True if the condition is met.
        timeout_message:           The message to include in the TimeoutError.
        timeout_seconds:           The maximum time to wait for the condition to be met.
        check_interval_seconds:    How long to wait inbetween `is_successful()` calls."""

    start_time = time.time()
    while True:
        if is_successful():
            break
        if (time.time() - start_time) > timeout_seconds:
            raise TimeoutError(timeout_message)
        time.sleep(check_interval_seconds)


_ISO_8601_PATTERN_1 = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?$")
_ISO_8601_PATTERN_2 = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$")
_ISO_8601_PATTERN_3 = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?(\+|-)\d{2}:\d{2}$")
_ISO_8601_PATTERN_4 = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?(\+|-)\d{4}$")
_ISO_8601_PATTERN_5 = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?(\+|-)\d{2}$")


def parse_iso_8601_datetime(s: str) -> datetime.datetime:
    """Parse a datetime string from various formats and return a datetime object.

    ISO 8601 supports time zones as `<time>Z`, `<time>±hh:mm`, `<time>±hhmm` and
    `<time>±hh`. However, only the second format is supported by `datetime.datetime.fromisoformat()`
    (`HH[:MM[:SS[.fff[fff]]]][+HH:MM[:SS[.ffffff]]]`).

    This function supports parsing alll ISO 8601 time formats."""

    if _ISO_8601_PATTERN_1.match(s) is not None:
        return datetime.datetime.fromisoformat(s)
    elif _ISO_8601_PATTERN_2.match(s) is not None:
        return datetime.datetime.fromisoformat(s[:-1] + "+00:00")
    elif _ISO_8601_PATTERN_3.match(s) is not None:
        return datetime.datetime.fromisoformat(s)
    elif _ISO_8601_PATTERN_4.match(s) is not None:
        return datetime.datetime.fromisoformat(s[:-2] + ":" + s[-2:])
    elif _ISO_8601_PATTERN_5.match(s) is not None:
        return datetime.datetime.fromisoformat(s + ":00")
    else:
        raise ValueError(f"Invalid datetime string: {s}")


def datetime_span_intersection(
    dt_span_1: tuple[datetime.datetime, datetime.datetime],
    dt_span_2: tuple[datetime.datetime, datetime.datetime],
) -> Optional[tuple[datetime.datetime, datetime.datetime]]:
    """Check if two datetime spans overlap.

    Args:
        dt_span_1: The first datetime span (start, end).
        dt_span_2: The second datetime span (start, end).

    Returns:
        The intersection of the two datetime spans or None if they do
        not overlap. Returns None if the intersection is a single point.
    """

    if dt_span_1[0] > dt_span_1[1]:
        raise ValueError("Invalid dt_span_1: start time is after end time")
    if dt_span_2[0] > dt_span_2[1]:
        raise ValueError("Invalid dt_span_2: start time is after end time")

    if dt_span_1[0] >= dt_span_2[1]:
        return None

    if dt_span_1[1] <= dt_span_2[0]:
        return None

    min_start = max(dt_span_1[0], dt_span_2[0])
    max_end = min(dt_span_1[1], dt_span_2[1])
    if min_start == max_end:
        return None
    return (min_start, max_end)


def date_span_intersection(
    d_span_1: tuple[datetime.date, datetime.date],
    d_span_2: tuple[datetime.date, datetime.date],
) -> Optional[tuple[datetime.date, datetime.date]]:
    """Check if two date spans overlap. This functions behaves
    differently from `datetime_span_intersection` in that it
    returns a single point as an intersection if the two date
    spans overlap at a single date.

    Args:
        d_span_1: The first date span (start, end).
        d_span_2: The second date span (start, end).

    Returns:
        The intersection of the two date spans or None if they do
        not overlap.
    """

    dt_intersection = datetime_span_intersection(
        (
            datetime.datetime.combine(d_span_1[0], datetime.time.min),
            datetime.datetime.combine(d_span_1[1], datetime.time.max),
        ),
        (
            datetime.datetime.combine(d_span_2[0], datetime.time.min),
            datetime.datetime.combine(d_span_2[1], datetime.time.max),
        ),
    )
    if dt_intersection is None:
        return None
    return (dt_intersection[0].date(), dt_intersection[1].date())


class ExponentialBackoff:
    """Exponential backoff e.g. when errors occur. First try again in 1 minute,
    then 4 minutes, then 15 minutes, etc.. Usage:

    ```python
    exponential_backoff = ExponentialBackoff(
        log_info=logger.info, buckets= [60, 240, 900, 3600, 14400]
    )

    while True:
        try:
            # do something that might fail
            exponential_backoff.reset()
        except Exception as e:
            logger.exception(e)
            exponential_backoff.sleep()
    ```"""

    def __init__(
        self,
        log_info: Optional[Callable[[str], None]] = None,
        buckets: list[int] = [60, 240, 900, 3600, 14400],
    ) -> None:
        """Create a new exponential backoff object.

        Args:
            log_info: The function to call when logging information.
            buckets:  The buckets to use for the exponential backoff.
        """

        self.buckets = buckets
        self.bucket_index = 0  # index of the next wait time bucket
        self.log_info = log_info

    def sleep(self, max_sleep_time: Optional[float] = None) -> float:
        """Wait and increase the wait time to the next bucket.

        Args:
            max_sleep_time: The maximum time to sleep. If None, no maximum is set.

        Returns:
            The amount of seconds waited.
        """

        sleep_seconds = float(self.buckets[self.bucket_index])
        if max_sleep_time is not None:
            sleep_seconds = min(sleep_seconds, max_sleep_time)

        minutes = math.floor(sleep_seconds / 60)
        seconds = sleep_seconds % 60

        message = "waiting for"
        if minutes > 0:
            message += f" {minutes} minute(s)"
            if seconds > 0:
                message += f" and"
        if seconds > 0:
            message += f" {round(seconds, 2)} second(s)"
        if self.log_info is not None:
            self.log_info(message)

        time.sleep(seconds)
        for i in range(1, minutes, 1):
            time.sleep(60)
            if self.log_info is not None:
                self.log_info(f"{minutes - i} minute(s) remaining")
        time.sleep(60)

        self.bucket_index = min(self.bucket_index + 1, len(self.buckets) - 1)
        return sleep_seconds

    def reset(self) -> None:
        """Reset the waiting period to the first bucket"""

        self.bucket_index = 0


@contextlib.contextmanager
def timed_section(label: str) -> Generator[None, None, None]:
    """Time a section of code and print the duration.
    Usage example:

    ```python
    with timed_section("my_section"):
        do_something()
    ```
    """

    start = time.time()
    yield
    end = time.time()
    print(f"{label}: {end - start:6.3f}s")


_JDN_BASE_DTS = {
    "JDN": datetime.datetime(2000, 1, 1, 12, 0, 0),
    "MJD": datetime.datetime(1858, 11, 17, 0, 0, 0),
    "MJD2K": datetime.datetime(2000, 1, 1, 0, 0, 0),
}
_JDN_BASE_NUMS = {
    "JDN": 2451545.0,
    "MJD": 0.0,
    "MJD2K": 0.0,
}


def datetime_to_julian_day_number(
    dt: datetime.datetime,
    variant: Literal["JDN", "MJD", "MJD2K"],
) -> float:
    """Convert a datetime to a Julian Day Number (JDN) or MJD/MJD2K.

    The Julian Day Number (JDN) is the continuous count of days since the beginning
    of the Julian Period on January 1, 4713 BC. THe modified variant MJD starts
    counting from November 17, 1858 at 00:00:00 UTC, and MJD2K starts counting
    from January 1, 2000 at 00:00:00 UTC.

    Args:
        dt: The datetime to convert.
        variant: The variant of the Julian Day Number ("JDN", "MJD", "MJD2K").

    Returns:
        The Julian Day Number as a float.
    """

    assert variant in _JDN_BASE_DTS, f"Invalid variant: {variant}"
    base_dt = _JDN_BASE_DTS[variant]
    base_num = _JDN_BASE_NUMS[variant]

    delta = dt - base_dt
    return base_num + delta.days + (delta.seconds + delta.microseconds / 1_000_000) / 86400.0


def julian_day_number_to_datetime(
    jdn: float,
    variant: Literal["JDN", "MJD", "MJD2K"],
) -> datetime.datetime:
    """Convert a Julian Day Number (JDN) or MJD/MJD2K to a datetime.

    The Julian Day Number (JDN) is the continuous count of days since the beginning
    of the Julian Period on January 1, 4713 BC. THe modified variant MJD starts
    counting from November 17, 1858 at 00:00:00 UTC, and MJD2K starts counting
    from January 1, 2000 at 00:00:00 UTC.

    This function was validated against
    https://ssd.jpl.nasa.gov/tools/jdc/#/cd

    Args:
        jdn: The Julian Day Number to convert.
        variant: The variant of the Julian Day Number ("JDN", "MJD", "MJD2K").

    Returns:
        The corresponding datetime.
    """

    assert variant in _JDN_BASE_DTS, f"Invalid variant: {variant}"
    base_dt = _JDN_BASE_DTS[variant]
    base_num = _JDN_BASE_NUMS[variant]

    delta_days = jdn - base_num
    delta_whole_days = int(delta_days)
    delta_fractional_day = delta_days - delta_whole_days

    delta_seconds = int(delta_fractional_day * 86400)
    delta_microseconds = int((delta_fractional_day * 86400 - delta_seconds) * 1_000_000)

    return base_dt + datetime.timedelta(
        days=delta_whole_days,
        seconds=delta_seconds,
        microseconds=delta_microseconds,
    )



"""Implements custom logging functionality, because the
standard logging module is hard to configure for special
cases.

Implements: `run_shell_command`, `CommandLineException`,
`get_hostname`, `get_commit_sha`, `change_file_permissions`"""

from typing import Callable, Literal, Optional
import os
import re
import subprocess


class CommandLineException(Exception):
    """Exception raised for errors in the command line."""

    def __init__(self, value: str, details: Optional[str] = None) -> None:
        self.value = value
        self.details = details
        Exception.__init__(self)

    def __str__(self) -> str:
        return repr(self.value)


def run_shell_command(
    command: str,
    working_directory: Optional[str] = None,
    executable: str = "/bin/bash",
    environment_variables: Optional[dict[str, str]] = None,
) -> str:
    """runs a shell command and raises a `CommandLineException`
    if the return code is not zero, returns the stdout. Uses
    `/bin/bash` by default.

    Args:
        command:               The command to run.
        working_directory:     The working directory for the command.
        executable:            The shell executable to use.
        environment_variables: A dictionary of environment variables to set for the command.

    Returns:
        The stdout of the command as a string."""

    env = os.environ.copy()
    if environment_variables is not None:
        env.update(environment_variables)

    p = subprocess.run(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=working_directory,
        env=env,
        executable=executable,
    )
    stdout = p.stdout.decode("utf-8", errors="replace").strip()
    stderr = p.stderr.decode("utf-8", errors="replace").strip()

    if p.returncode != 0:
        raise CommandLineException(
            f"command '{command}' failed with exit code {p.returncode}",
            details=f"\nstderr:\n{stderr}\nstout:\n{stdout}",
        )

    return stdout


def get_hostname() -> str:
    """returns the hostname of the device, removes network
    postfix (`somename.local`) if present. Only works reliably,
    when the hostname doesn't contain a dot."""

    raw = run_shell_command("hostname")
    return (raw.split(".")[0]) if ("." in raw) else raw


def get_commit_sha(variant: Literal["short", "long"] = "short") -> Optional[str]:
    """Get the current commit sha of the repository. Returns
    `None` if there is not git repository in any parent directory.

    Args:
        variant:  "short" or "long" to specify the length of the sha.

    Returns:
        The commit sha as a string, or `None` if there is no git
        repository in the parent directories."""

    try:
        p = subprocess.run(
            ["git", "rev-parse", "--verify", "HEAD"] + (["--short"] if variant == "short" else []),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        stdout = p.stdout.decode().strip("\n ")
        assert p.returncode == 0
        assert "fatal: not a git repository" not in stdout
        assert len(stdout) > 0
    except:
        return None

    return stdout


_permission_string_pattern = re.compile(r"^((r|-)(w|-)(x|-)){3}$")


def change_file_permissions(file_path: str, permission_string: str) -> None:
    """Change a file's system permissions.

    Example permission_strings: `--x------`, `rwxr-xr-x`, `rw-r--r--`.

    Args:
        file_path:         The path to the file.
        permission_string: The new permission string."""

    if os.name != "posix":
        raise OSError("change_file_permissions is only supported on Unix systems")

    assert _permission_string_pattern.match(permission_string), "Invalid permission string"

    permission_str_to_bit: Callable[[str], int] = lambda p: sum(
        [int(c) for c in p.replace("r", "4").replace("w", "2").replace("x", "1").replace("-", "0")]
    )
    os.chmod(
        file_path,
        64 * permission_str_to_bit(permission_string[:3])
        + 8 * permission_str_to_bit(permission_string[3:6])
        + permission_str_to_bit(permission_string[6:]),
    )



"""A thin wrapper over the netCDF4 library to make working with NetCDF files easier.

Implements: `NetCDFFile`, `remove_elements_from_netcdf_file`, `compress_netcdf_file`.

This requires you to install this utils library with the optional `netcdf` dependencies:

```bash
pip install "tum_esm_utils[netcdf]"
# or
pdm add "tum_esm_utils[netcdf]"
```"""

from typing import Any, Literal, Optional
import os
import netCDF4 as nc


class NetCDFFile:
    def __init__(
        self,
        filepath: str,
        parallel: bool = False,
        diskless: bool = False,
        mode: Literal["r", "w", "r+", "a", "x", "rs", "ws", "r+s", "as"] = "r",
    ) -> None:
        """A simple wrapper around netCDF4.Dataset to make the interaction with NetCDF files easier.

        If writing to a new file, it will first write to the filepath+ ".tmp" and  rename it to the final
        filepath when closing the file. This ensures that the final filepath will only exist if the file
        was written completely. In append mode, the filepath is not changes."""

        extension = filepath.split(".")[-1]
        self.tmp_filepath = filepath[: -(len(extension) + 1)] + f".tmp.{extension}"
        self.filepath = filepath
        self.mode = mode

        if mode == "w" and os.path.isfile(self.tmp_filepath):
            os.remove(self.tmp_filepath)

        self.ds = nc.Dataset(
            self.tmp_filepath if mode == "w" else self.filepath,
            mode=mode,
            format="NETCDF4",
            parallel=parallel,
            diskless=diskless,
            persist=True,
        )
        self.dimensions: dict[str, nc.Dimension] = {}
        self.variables: dict[str, nc.Variable[Any]] = {}
        self.attributes: dict[str, str] = {}

        if mode != "w":
            for dim_name, dim in self.ds.dimensions.items():
                self.dimensions[dim_name] = dim
            for var_name, var in self.ds.variables.items():
                self.variables[var_name] = var
            for attr_name in self.ds.ncattrs():
                self.attributes[attr_name] = self.ds.getncattr(attr_name)

    def create_dimension(self, name: str, size: int) -> None:
        """Create a new dimension in the NetCDF file.

        Raises:
            ValueError: If the dimension already exists
            RuntimeError: If the NetCDF file is not opened in write mode."""

        if self.mode == "r":
            raise RuntimeError("Cannot create dimension in read-only mode")

        if name in self.dimensions:
            raise ValueError(f"Dimension {name} already exists in the NetCDF file")

        d = self.ds.createDimension(name, size)
        assert isinstance(d, nc.Dimension)
        self.dimensions[name] = d

    def create_variable(
        self,
        name: str,
        dimensions: tuple[nc.Dimension | str, ...],
        units: str,
        long_name: Optional[str] = None,
        description: Optional[str] = None,
        fill_value: Optional[float | int] = None,
        chunk_dimensions: list[str] = [],
        datatype: Literal["f4", "f8", "i4", "i8"] = "f4",
        zlib: bool = True,
        compression_level: int = 2,
    ) -> None:
        """Create a new variable in the NetCDF file.

        Raises:
            ValueError: If the variable already exists or if a dimension is not found.
            RuntimeError: If the NetCDF file is not opened in write mode."""

        if self.mode == "r":
            raise RuntimeError("Cannot create dimension in read-only mode")

        if name in self.variables:
            raise ValueError(f"Variable {name} already exists in the NetCDF file")

        object_dimensions: list[nc.Dimension] = []
        for dimension in dimensions:
            if isinstance(dimension, str):
                if dimension not in self.dimensions:
                    raise ValueError(f"Dimension {dimension} not found in the NetCDF file")
                object_dimensions.append(self.dimensions[dimension])
            else:
                if dimension.name not in self.dimensions:
                    raise ValueError(f"Dimension {dimension.name} not found in the NetCDF file")
                object_dimensions.append(dimension)

        chunk_sizes = [dimension.size for dimension in object_dimensions]
        for i, dimension in enumerate(object_dimensions):
            if dimension.name in chunk_dimensions:
                chunk_sizes[i] = 1

        var: Any = self.ds.createVariable(  # pyright: ignore[reportUnknownMemberType,reportUnknownVariableType]
            name,
            datatype=datatype,
            dimensions=object_dimensions,
            zlib=zlib and ((len(dimensions) > 1) or (name != object_dimensions[0].name)),
            complevel=compression_level,  # type: ignore
            fill_value=fill_value,
            chunksizes=chunk_sizes if len(chunk_dimensions) > 0 else None,
        )
        var.units = units
        if long_name is not None:
            var.long_name = long_name
        if description is not None:
            var.description = description
        self.variables[name] = var

    def import_dimension(
        self,
        dimension: nc.Dimension,
        new_name: Optional[str] = None,
    ) -> None:
        """Import a dimension from another NetCDF file.

        Raises:
            ValueError: If the dimension already exists.
            RuntimeError: If the NetCDF file is not opened in write mode."""

        if self.mode == "r":
            raise RuntimeError("Cannot import dimension in read-only mode")

        if dimension.name in self.dimensions:
            raise ValueError(f"Dimension {dimension.name} already exists in the NetCDF file")
        self.create_dimension(dimension.name if new_name is None else new_name, dimension.size)

    def import_variable(
        self,
        variable: "nc.Variable[Any]",
        new_name: Optional[str] = None,
        zlib: bool = True,
        compression_level: int = 2,
    ) -> None:
        """Import a variable from another NetCDF file.

        Raises:
            ValueError: If the variable already exists.
            RuntimeError: If the NetCDF file is not opened in write mode."""

        if self.mode == "r":
            raise RuntimeError("Cannot import variable in read-only mode")

        if variable.name in self.variables:
            raise ValueError(f"Variable {variable.name} already exists in the NetCDF file")
        name = variable.name if new_name is None else new_name
        self.create_variable(
            name=name,
            dimensions=variable.dimensions,
            units=str(variable.units),
            long_name=variable.long_name if hasattr(variable, "long_name") else None,  # pyright: ignore[reportUnknownArgumentType]
            description=variable.description if hasattr(variable, "description") else None,  # pyright: ignore[reportUnknownArgumentType]
            fill_value=float(variable.get_fill_value()),
            zlib=zlib,
            compression_level=compression_level,
        )
        self.variables[name][:] = variable[:]

    def add_attribute(self, key: str, value: str, allow_overwrite: bool = False) -> None:
        """Add a global attribute to the NetCDF file.

        Raises:
            ValueError: If the attribute already exists and `allow_overwrite` is False.
            RuntimeError: If the NetCDF file is not opened in write mode."""

        if self.mode == "r":
            raise RuntimeError("Cannot add attribute in read-only mode")

        if (not allow_overwrite) and (key in self.attributes):
            raise ValueError(f"Attribute {key} already exists in the NetCDF file")
        self.attributes[key] = value
        self.ds.setncattr(key, value)

    def close(self) -> None:
        """Close the NetCDF file, possibly renaming the temporary file to the final filepath."""

        self.ds.close()
        if self.mode == "w":
            if os.path.isfile(self.filepath):
                os.remove(self.filepath)
            os.rename(self.tmp_filepath, self.filepath)

        del self

    def __getitem__(self, key: str) -> "nc.Variable[Any]":
        """Get a variable from the NetCDF file."""
        return self.variables[key]


def remove_elements_from_netcdf_file(
    source_filepath: str,
    destination_filepath: str,
    variables_to_remove: list[str] = [],
    dimensions_to_remove: list[str] = [],
    attributes_to_remove: list[str] = [],
    compression_level: int = 2,
) -> None:
    """Create a new NetCDF file by copying an existing one, but removing specified variables, dimensions, and attributes. This is useful because NetCDF4 does not support removing elements from an existing file.

    Raises:
        FileNotFoundError: If the source file does not exist.
        FileExistsError: If the destination file already exists.
    """

    if not os.path.isfile(source_filepath):
        raise FileNotFoundError(f"Source file {source_filepath} does not exist.")
    if os.path.isfile(destination_filepath):
        raise FileExistsError(f"Destination file {destination_filepath} already exists.")

    src_nc = NetCDFFile(source_filepath, mode="r")
    dest_nc = NetCDFFile(destination_filepath, mode="w")

    # check that no variable depends on a dimension to be removed
    vars = [v for v in src_nc.variables.values() if v.name not in variables_to_remove]
    for var in vars:
        for dim_name in var.dimensions:
            if dim_name in dimensions_to_remove:
                raise ValueError(
                    f"Cannot remove dimension {dim_name} because it is used by variable {var.name}."
                )

    # Copy dimensions
    for dim_name, dim in src_nc.dimensions.items():
        if dim_name not in dimensions_to_remove:
            dest_nc.import_dimension(dim)

    # Copy variables
    for var_name, var in src_nc.variables.items():
        if var_name not in variables_to_remove:
            dest_nc.import_variable(var, compression_level=compression_level)

    # Copy attributes
    for attr_name, attr_value in src_nc.attributes.items():
        if attr_name not in attributes_to_remove:
            dest_nc.add_attribute(attr_name, attr_value)

    src_nc.close()
    dest_nc.close()


def compress_netcdf_file(
    source_filepath: str,
    destination_filepath: str,
    compression_level: int = 2,
) -> None:
    """Compress an existing NetCDF file by creating a new one with the specified compression level. This is useful because some NetCDF4 files given to you might not be (very well) compressed.

    Raises:
        FileNotFoundError: If the source file does not exist.
        FileExistsError: If the destination file already exists.
    """

    remove_elements_from_netcdf_file(
        source_filepath,
        destination_filepath,
        compression_level=compression_level,
    )



"""Functions for interacting with EM27 interferograms.

Implements: `detect_corrupt_opus_files`, `load_proffast2_result`.

This requires you to install this utils library with the optional `em27` dependency:

```bash
pip install "tum_esm_utils[em27]"
# or
pdm add "tum_esm_utils[em27]"
```"""

from __future__ import annotations
from typing import Literal
from typing_extensions import deprecated
import os
import subprocess
import filelock
import tum_esm_utils.text
import polars as pl
from tailwind_colors import TAILWIND_COLORS_HEX as TCH

_PARSER_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "opus_file_validator")


def _compile_fortran_code(
    silent: bool = True,
    fortran_compiler: Literal["gfortran", "gfortran-9"] = "gfortran",
    force_recompile: bool = False,
) -> None:
    assert os.name == "posix"

    if force_recompile or (not os.path.isfile(os.path.join(_PARSER_DIR, "opus_file_validator"))):
        if not silent:
            print("compiling fortran code")

        command = (
            f"{fortran_compiler} -nocpp -O3 -o ./opus_file_validator "
            + f"glob_prepro6.F90 glob_OPUSparms6.F90 opus_file_validator.F90"
        )
        p = subprocess.run(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=_PARSER_DIR,
            env=os.environ.copy(),
        )
        if p.returncode != 0:
            stdout = p.stdout.decode("utf-8", errors="replace").strip()
            stderr = p.stderr.decode("utf-8", errors="replace").strip()
            raise Exception(
                f"command '{command}' failed with exit code {p.returncode}, "
                + f"stderr: {stderr}, stout:{stdout}",
            )


def detect_corrupt_opus_files(
    ifg_directory: str,
    silent: bool = True,
    fortran_compiler: Literal["gfortran", "gfortran-9"] = "gfortran",
    force_recompile: bool = False,
) -> dict[str, list[str]]:
    """Returns dict[filename, list[error_messages]] for all
    corrupt opus files in the given directory.

    It will compile the fortran code using a given compiler
    to perform this task. The fortran code is derived from
    the preprocess source code of Proffast 2
    (https://www.imk-asf.kit.edu/english/3225.php). We use
    it because the retrieval using Proffast 2 will fail if
    there are corrupt interferograms in the input.

    Args:
        ifg_directory:     The directory containing the interferograms.
        silent:            If set to False, print additional information.
        fortran_compiler:  The fortran compiler to use.
        force_recompile:   If set to True, the fortran code will be recompiled.

    Returns:
        A dictionary containing corrupt filenames as keys and a list of error
        messages as values."""

    if os.name != "posix":
        raise OSError("detect_corrupt_opus_files is only supported on Unix systems")

    # compiling the fortran code in a semaphore
    with filelock.FileLock(
        os.path.join(_PARSER_DIR, "opus_file_validator.lock"),
        timeout=30,
    ):
        _compile_fortran_code(
            silent=silent, fortran_compiler=fortran_compiler, force_recompile=force_recompile
        )

    # list directory files
    filepaths = list(
        sorted(
            [
                fp
                for fp in [f"{ifg_directory}/{x}" for x in os.listdir(ifg_directory)]
                if os.path.isfile(fp)
            ]
        )
    )

    # write input file for parser in a semaphore
    with filelock.FileLock(f"{_PARSER_DIR}/opus_file_validator.lock"):
        random_id = tum_esm_utils.text.get_random_string(
            10,
            forbidden=[
                f.replace("opus_file_validator.inp.", "")
                for f in os.listdir(_PARSER_DIR)
                if f.startswith("opus_file_validator.inp.")
            ],
        )
        input_file_path = f"{_PARSER_DIR}/opus_file_validator.inp.{random_id}"

        with open(f"{_PARSER_DIR}/opus_file_validator.template.inp", "r") as f:
            template_content = f.read()
        with open(input_file_path, "w") as f:
            f.write(template_content.replace("%IFG_LIST%", "\n".join(filepaths)))

    # run the parser
    process = subprocess.run(
        ["./opus_file_validator", input_file_path],
        cwd=_PARSER_DIR,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    os.remove(input_file_path)
    stdout = process.stdout.decode()
    stderr = process.stderr.decode()

    if not process.returncode == 0:
        raise RuntimeError(
            f"Opus File Parser failed with exit code {process.returncode}, "
            + f"stderr: {stderr}, stdout: {stdout}",
        )

    # locate the block of verification results
    if (stdout.count("--- Start verifying file integrities ---") != 1) or (
        stdout.count("--- Done verifying file integrities ---") != 1
    ):
        raise Exception("This is a bug in the `tum_esm_utils` library")
    verification_block = (
        stdout.split("--- Start verifying file integrities ---")[1]
        .split("--- Done verifying file integrities ---")[0]
        .strip("\t\n ")
    )

    # parse the verification results
    results: dict[str, list[str]] = {}
    checked_files: set[str] = set(filepaths)
    if "\n\n" in verification_block:
        file_verification_blocks = verification_block.split("\n\n")
        for block in file_verification_blocks:
            lines = block.split("\n")
            is_corrupt = len(lines) > 2
            filepath = lines[0].split('"')[1]
            if is_corrupt:
                results[os.path.basename(filepath)] = lines[1:-1]
            checked_files.remove(filepath)

    # every file not mentioned in the verification results failed during reading it
    for filepath in checked_files:
        results[os.path.basename(filepath)] = ["File not even readible by the parser"]

    # save the raw output for debugging purposes
    with open(os.path.join(_PARSER_DIR, "output.txt"), "w") as f:
        f.write(stdout)

    return results


@deprecated(
    "This will be removed in the next breaking release. Please use "
    + "the identical function `detect_corrupt_opus_files` instead."
)
def detect_corrupt_ifgs(
    ifg_directory: str,
    silent: bool = True,
    fortran_compiler: Literal["gfortran", "gfortran-9"] = "gfortran",
    force_recompile: bool = False,
) -> dict[str, list[str]]:
    """Returns dict[filename, list[error_messages]] for all
    corrupt opus files in the given directory.

    It will compile the fortran code using a given compiler
    to perform this task. The fortran code is derived from
    the preprocess source code of Proffast 2
    (https://www.imk-asf.kit.edu/english/3225.php). We use
    it because the retrieval using Proffast 2 will fail if
    there are corrupt interferograms in the input.

    Args:
        ifg_directory:     The directory containing the interferograms.
        silent:            If set to False, print additional information.
        fortran_compiler:  The fortran compiler to use.
        force_recompile:   If set to True, the fortran code will be recompiled.

    Returns:
        A dictionary containing corrupt filenames as keys and a list of error
        messages as values."""
    return detect_corrupt_opus_files(
        ifg_directory=ifg_directory,
        silent=silent,
        fortran_compiler=fortran_compiler,
        force_recompile=force_recompile,
    )


def load_proffast2_result(path: str) -> pl.DataFrame:
    """Loads the output of Proffast 2 into a polars DataFrame.

    Args:
        path: The path to the Proffast 2 output file.

    Returns:
        A polars DataFrame containing all columns.
    """

    # fmt: off
    column_names = [
        "JulianDate", "UTtimeh", "gndP", "gndT", "latdeg", "londeg", "altim",
        "appSZA", "azimuth", "XH2O", "XAIR", "XCO2", "XCH4", "XCO2_STR", "XCO",
        "XCH4_S5P", "H2O", "O2", "CO2", "CH4", "CO", "CH4_S5P"
    ]
    # fmt: on
    df = pl.read_csv(
        path,
        has_header=True,
        separator=",",
        schema_overrides={
            "UTC": pl.Datetime,
            " LocalTime": pl.Utf8,
            " spectrum": pl.Utf8,
            **{f" {cn}": pl.Float32 for cn in column_names},
        },
    ).drop(" JulianDate", " UTtimeh")
    return df.rename(
        {
            " LocalTime": "LocalTime",
            " spectrum": "spectrum",
            **{f" {cn}": cn for cn in column_names if f" {cn}" in df.columns},
        }
    ).with_columns(
        pl.col("LocalTime").str.strptime(dtype=pl.Datetime, format=" %Y-%m-%d %H:%M:%S"),
    )


SERIAL_NUMBERS: dict[str, int] = {
    "ma": 61,
    "mb": 86,
    "mc": 115,
    "md": 116,
    "me": 117,
    "kc": 83,
}
"""The serial numbers of the EM27 devices."""

COLORS: dict[str, str] = {
    "ma": TCH.YELLOW_500,
    "mb": TCH.ORANGE_500,
    "mc": TCH.PINK_500,
    "md": TCH.VIOLET_500,
    "me": TCH.BLUE_500,
    "kc": TCH.GREEN_500,
}
"""Colors recommended for plotting the EM27 data."""

COLORS_LIGHT: dict[str, str] = {
    "ma": TCH.YELLOW_400,
    "mb": TCH.ORANGE_400,
    "mc": TCH.PINK_400,
    "md": TCH.VIOLET_400,
    "me": TCH.BLUE_400,
    "kc": TCH.GREEN_400,
}
"""Lighter colors recommended for plotting the EM27 data."""

COLORS_DARK: dict[str, str] = {
    "ma": TCH.YELLOW_600,
    "mb": TCH.ORANGE_600,
    "mc": TCH.PINK_600,
    "md": TCH.VIOLET_600,
    "me": TCH.BLUE_600,
    "kc": TCH.GREEN_600,
}
"""Darker colors recommended for plotting the EM27 data."""

PROFFAST_MULTIPLIERS: dict[str, float] = {
    "XCO2": 1,
    "XCH4": 1000,
    "XCO": 1000,
    "XAIR": 1,
    "XH2O": 1,
    "XAIR": 1,
    "XCO2": 1,
    "XCO2_STR": 1,
    "XCH4": 1000,
    "XCO": 1000,
    "XCH4_S5P": 1000,
    "XAIR": 1,
    "H2O": 1 / (6.022 * 10e23),
    "O2": 1 / (6.022 * 10e23),
    "CO2": 1 / (6.022 * 10e23),
    "CH4": 1 / (6.022 * 10e23),
    "CO": 1 / (6.022 * 10e23),
    "CH4_S5P": 1 / (6.022 * 10e23),
}
"""Multiplication factors for the EM27 data retrieved using Proffast to bring the data in a common unit."""

PROFFAST_UNITS: dict[str, str] = {
    "XH2O": "ppm",
    "XAIR": "ppm",
    "XCO2": "ppm",
    "XCO2_STR": "ppm",
    "XCH4": "ppb",
    "XCO": "ppb",
    "XCH4_S5P": "ppb",
    "XAIR": "-",
    "H2O": "mol",
    "O2": "mol",
    "CO2": "mol",
    "CH4": "mol",
    "CO": "mol",
    "CH4_S5P": "mol",
}
"""Units for the EM27 data retrieved using Proffast after applying the multiplication factor."""



"""Utilities for interacting with OPUS files."""

from __future__ import annotations
from typing import Optional, TypeVar
import warnings
import numpy as np
import numpy.typing as npt
import io
import struct
from . import types

# Define constants

FILE_MAGIC = b"\n\n\xfe\xfe"  # magic sequence at beginning of Opus file
# Definitions for directory block type bit patterns

# Parameter type flag
DBSPARM = 4  # bit shift
DBMPARM = 63  # bit mask
DBFPARM: list[str] = [
    "",  # undefined
    "DBTDSTAT",  # data status parameter (DataParameters )
    "DBTINSTR",  # instrument status parameters
    "DBTAQPAR",  # standard acquisition parameters
    "DBTFTPAR",  # FT-Parameters
    "DBTPLTPAR",  # plot- and display parameters
    "DBTPRCPAR",  # processing parameters (Optic parameters)
    "DBTGCPAR",  # GC-parameters
    "DBTLIBPAR",  # library search parameters
    "DBTCOMPAR",  # communication parameters
    "DBTORGPAR",  # sample origin parameter (Sample Parameters)
] + [f"DBTPARM{i}" for i in range(11, DBMPARM + 1)]
assert len(DBFPARM) == DBMPARM + 1


def read_opus_header(f: io.BufferedReader) -> types.OpusHeader:
    """Read OPUS file header from a file."""

    magic_sequence = f.read(4)
    if magic_sequence != FILE_MAGIC:
        raise RuntimeError(f'Magic sequence not found, found {magic_sequence.decode()}" instead')
    x = struct.unpack("<dIII", f.read(struct.calcsize("<dIII")))
    assert len(x) >= 4
    return types.OpusHeader(
        version=x[0],
        dir_pointer=x[1],
        max_dir_size=x[2],
        dir_size=x[3],
    )


def read_opus_dir_entry(f: io.BufferedReader) -> types.OpusDirectoryEntry:
    """Read a single OPUS directory entry object from file at current position."""

    fmt = "<III"
    x = struct.unpack(fmt, f.read(struct.calcsize(fmt)))
    assert len(x) == 3
    block_type = x[0]
    block_length = x[1]
    block_pointer = x[2]
    assert isinstance(block_type, int)
    assert isinstance(block_length, int)
    assert isinstance(block_pointer, int)

    return types.OpusDirectoryEntry(
        block_type=block_type,
        block_length=block_length,
        block_pointer=block_pointer,
        block_category=DBFPARM[(block_type >> DBSPARM) & DBMPARM],
    )


T = TypeVar("T", types.OpusParameterBlock, types.OpusDataBlock)


def read_opus_block(
    f: io.BufferedReader,
    opus_directory_entry: types.OpusDirectoryEntry,
    expected_block_type: type[T],
) -> T:
    f.seek(opus_directory_entry.block_pointer)
    raw_data = f.read(4 * opus_directory_entry.block_length)

    block: types.OpusParameterBlock | types.OpusDataBlock
    if opus_directory_entry.block_category == "":
        block = types.OpusDataBlock(raw_data=raw_data)
        if not isinstance(block, expected_block_type):
            raise RuntimeError(f"Expected a {expected_block_type}, but got a {type(block)}")
        return block

    data: dict[str, Optional[int | float | str]] = {}
    data_types: dict[str, int] = {}
    parameter_order: list[str] = []

    offset = 0  # Byte offset from beginning of block
    fmt1 = "<4shh"  # Initial format string: 4-byte string, 2 short int (little endian)
    while offset <= (len(raw_data) - struct.calcsize(fmt1)):
        # rs: seserved space in 16 bit units
        (pname, ptype, rs) = struct.unpack_from(fmt1, raw_data, offset)
        assert isinstance(pname, bytes)
        assert isinstance(ptype, int)
        assert isinstance(rs, int)

        parameter_name = pname[:-1].decode("utf-8")  # Remove '\0' terminator
        offset += struct.calcsize(fmt1)
        value: Optional[int | float | str] = None

        # Try to read the following values
        if rs > 0:
            # INT32 (little endian)
            if ptype == 0:
                fmt2 = "<i"
                raw_value = struct.unpack_from("<i", raw_data, offset)
                assert isinstance(raw_value[0], int)
                value = raw_value[0]

            # REAL64 (little endian)
            elif ptype == 1:
                fmt2 = "<d"
                raw_value = struct.unpack_from(fmt2, raw_data, offset)
                assert isinstance(raw_value[0], float)
                value = raw_value[0]

            # STRING, ENUM or SENUM
            else:
                fmt2 = ("<%ds") % (2 * rs)
                raw_value = struct.unpack_from(fmt2, raw_data, offset)
                assert isinstance(raw_value[0], bytes)
                value = raw_value[0][:-1].decode("utf-8").replace("\x00", "")

            offset += struct.calcsize(fmt2)

        # Store decoded parameter value
        data[parameter_name] = value
        data_types[parameter_name] = ptype
        parameter_order.append(parameter_name)

    block = types.OpusParameterBlock(
        data=data,
        data_types=data_types,
        parameter_order=parameter_order,
        raw_data=raw_data,
    )
    if not isinstance(block, expected_block_type):
        raise RuntimeError(f"Expected a {expected_block_type}, but got a {type(block)}")
    return block


def read_interferogram(
    f: io.BufferedReader,
    channel_parameters: list[types.OpusChannelParameters],
    ifg_opus_dirs: list[types.OpusDirectoryEntry],
    read_all_channels: bool = True,
) -> npt.NDArray[np.float64]:
    if len(channel_parameters) < len(ifg_opus_dirs):
        raise RuntimeError("There are fewer channel parameter blocks than interferogram blocks!")

    if len(channel_parameters) == 0:
        raise RuntimeError("No channel parameters found!")

    if not read_all_channels:
        channel_parameters = channel_parameters[:1]
        ifg_opus_dirs = ifg_opus_dirs[:1]

    spectrum_length = channel_parameters[0].spectrum["NPT"]
    for channel_parameter in channel_parameters[1:]:
        if channel_parameter.spectrum["NPT"] != spectrum_length:
            warnings.warn(
                "Channel parameters indicate different lengths. Using the first channel's length."
            )

    if len(ifg_opus_dirs) not in [1, 2]:
        raise RuntimeError(f"Invalid number of interferogram blocks found: {len(ifg_opus_dirs)}")

    for ifg_opus_dir in ifg_opus_dirs:
        if ifg_opus_dir.block_length != spectrum_length:
            raise RuntimeError(
                f"Interferogram block has length {ifg_opus_dir.block_length}, "
                f"but expected {spectrum_length}"
            )

    full_ifg = np.zeros(
        shape=(len(ifg_opus_dirs), spectrum_length),
        dtype=np.float64,
    )
    for channel_index in range(len(ifg_opus_dirs)):
        channel_ifg = np.multiply(
            np.array(
                struct.unpack_from(
                    f"<{spectrum_length}f",
                    read_opus_block(f, ifg_opus_dirs[0], types.OpusDataBlock).raw_data,
                    offset=0,
                )
            ),
            channel_parameters[channel_index].spectrum["CSF"],
        )
        full_ifg[channel_index, :] = channel_ifg.copy()

    return full_ifg



"""Functions to rebin binned data poins

Implements: `rebin_1d`, `rebin_2d`.

This requires you to install this utils library with the optional `modeling` dependency:

```bash
pip install "tum_esm_utils[modeling]"
# or
pdm add "tum_esm_utils[modeling]"
```"""

from typing import Any
import numpy as np


def _rebin_first_dimension(
    arr: np.ndarray[Any, Any],
    new_bin_count: int,
) -> np.ndarray[Any, Any]:
    """Internal rebinning function."""

    old_bin_count = len(arr)
    new_bins = np.zeros(shape=(new_bin_count, *arr.shape[1:]), dtype=np.float64)
    scale = old_bin_count / new_bin_count
    for i in range(new_bin_count):
        start = i * scale
        end = (i + 1) * scale
        left = int(np.floor(start))
        right = int(np.floor(end))

        if left == right:
            new_bins[i] += arr[left] * (end - start)
        else:
            new_bins[i] += arr[left] * (left + 1 - start)
            for j in range(left + 1, right):
                new_bins[i] += arr[j]
            if right < old_bin_count:
                new_bins[i] += arr[right] * (end - right)
    return new_bins


def rebin_1d(
    arr: np.ndarray[Any, Any],
    new_bin_count: int,
) -> np.ndarray[Any, Any]:
    """Rebins a 1D array to a new number of bins."""
    if len(arr.shape) != 1:
        raise ValueError("Input array must be 1D.")
    return _rebin_first_dimension(arr, new_bin_count)


def rebin_2d(
    arr: np.ndarray[Any, Any],
    new_x_bins: int,
    new_y_bins: int,
) -> np.ndarray[Any, Any]:
    """Rebins a 2D array to new number of bins in x and y dimensions."""
    return _rebin_first_dimension(_rebin_first_dimension(arr.T, new_x_bins).T, new_y_bins)



"""Python utilities by the Professorship of Environmental
Sensing and Modeling at the Technical University of Munich.

GitHub Repository https://github.com/tum-esm/utils
Documentation: https://tum-esm-utils.netlify.app
PyPI: https://pypi.org/project/tum-esm-utils

(Optional) Explicit Imports:

By setting the environment variable `TUM_ESM_UTILS_EXPLICIT_IMPORTS=1`, the
package disables automatic submodule imports. This means you cannot import the
whole package and access submodules directly (e.g., `tum_esm_utils.code` will
not be available after `import tum_esm_utils`). Instead, you must explicitly
import each submodule, e.g. `from tum_esm_utils import code` or
`import tum_esm_utils.code`.

This reduces the import time of the package by up to 60 times
"""

import os

if os.getenv("TUM_ESM_UTILS_EXPLICIT_IMPORTS") != "1":
    from . import (
        code as code,
        datastructures as datastructures,
        decorators as decorators,
        files as files,
        mathematics as mathematics,
        processes as processes,
        shell as shell,
        sqlitelock as sqlitelock,
        system as system,
        text as text,
        timing as timing,
        validators as validators,
    )

    # ignore import errors from the following submodules
    # because they requires extras to be installed

    # requires extra "em27"
    try:
        from . import em27 as em27
    except ImportError:
        pass

    # requires extra "plotting"
    try:
        from . import plotting as plotting
    except ImportError:
        pass

    # requires extra "opus"
    try:
        from . import opus as opus
    except ImportError:
        pass

    # requires extra "modeling"
    try:
        from . import column as column
    except ImportError:
        pass

    # requires extra "modeling"
    try:
        from . import rebinning as rebinning
    except ImportError:
        pass

    # requires extra "netcdf"
    try:
        from . import netcdf as netcdf
    except ImportError:
        pass



"""Types for interacting with OPUS files."""

from __future__ import annotations
from typing import Any, Literal, Optional
import math
import datetime
import pydantic
import tum_esm_utils.timing


class OpusHeader(pydantic.BaseModel):
    """Object that contains all the information in the OPUS file header block."""

    version: int = pydantic.Field(..., ge=0, description="File version")
    dir_pointer: int = pydantic.Field(
        ..., ge=0, description="File pointer to start of OPUS directory"
    )
    max_dir_size: int = pydantic.Field(..., ge=0, description="Maximum number of directory entries")
    dir_size: int = pydantic.Field(..., ge=0, description="Used number of OPUS directory entries")


class OpusParameterBlock(pydantic.BaseModel):
    data: dict[str, Optional[str | float | int]]
    data_types: dict[str, int]
    parameter_order: list[str]
    raw_data: bytes


class OpusDataBlock(pydantic.BaseModel):
    raw_data: bytes


class OpusDirectoryEntry(pydantic.BaseModel):
    """Object that contains all the information in an OPUS directory entry."""

    block_type: int = pydantic.Field(..., description="encoded block type (32 bit mask)")
    block_length: int = pydantic.Field(..., description="block length (32-bit words)")
    block_pointer: int = pydantic.Field(
        ..., description="file pointer to beginning of data/parameter block (bytes)"
    )
    block_category: str = pydantic.Field(
        ..., description="decoded parameter type flag (values from DBSPARM)"
    )


class OpusChannelParameters(pydantic.BaseModel):
    spectrum: dict[str, Any] = pydantic.Field(..., validation_alias="DBTDSTAT")
    instrument: dict[str, Any] = pydantic.Field(..., validation_alias="DBTINSTR")
    acquisition: dict[str, Any] = pydantic.Field(..., validation_alias="DBTAQPAR")
    optics: dict[str, Any] = pydantic.Field(..., validation_alias="DBTPRCPAR")
    sample: dict[str, Any] = pydantic.Field(..., validation_alias="DBTORGPAR")
    fourier_transform: dict[str, Any] = pydantic.Field(..., validation_alias="DBTFTPAR")

    def parse_measurement_datetime(
        self,
        measurement_timestamp_mode: Literal["start", "end"] = "start",
    ) -> datetime.datetime:
        DAT, TIM, DUR = self.spectrum["DAT"], self.spectrum["TIM"], self.instrument["DUR"]
        seconds = float(TIM.split(":")[2].split(" ")[0])
        utc_offset = tum_esm_utils.timing.parse_timezone_string(TIM.split(" ")[-1].strip("()"))
        dt = datetime.datetime(
            year=int(DAT.split("/")[2]),
            month=int(DAT.split("/")[1]),
            day=int(DAT.split("/")[0]),
            hour=int(TIM.split(":")[0]),
            minute=int(TIM.split(":")[1]),
            second=math.floor(seconds),
            microsecond=round((seconds % 1) * 1_000_000),
            tzinfo=datetime.timezone.utc,
        ) - datetime.timedelta(hours=utc_offset)
        if measurement_timestamp_mode == "start":
            dt = dt + datetime.timedelta(seconds=DUR / 2)
        else:
            dt = dt - datetime.timedelta(seconds=DUR / 2)
        return dt



from typing import Any, Optional
import os
import time
import sqlite3


class SQLiteLock:
    """A file lock based on SQLite transactions.

    The alternative `filelock` package tends to deadlock on our low-spec-CPU
    windows machines. The package `portalocker` uses the `pywin32` package
    which I am not a big fan of due to its documentation and testing quality.

    Usage example:

    ```python
    lock = tum_esm_utils.sqlitelock.SQLiteLock("sqlitelock.lock", timeout=5)

    try:
        with lock:
            # critical section
            pass
    except TimeoutError:
        # could not be acquired within 5 seconds
        pass
    ```

    This function is tested on Windows, Linux."""

    def __init__(
        self,
        filepath: str = "sqlitelock.lock",
        timeout: float = 10,
        poll_interval: float = 0.1,
    ) -> None:
        """Initialize the SqliteFileLock.

        Args:
            filepath: The path to the SQLite database file used for locking.
            timeout: The maximum time to wait for acquiring the lock in seconds.
            poll_interval: The interval between lock acquisition attempts in seconds.
        """

        self.timeout = timeout
        self.poll_interval = poll_interval
        self._is_locked: bool = False

        # Ensure directory exists
        dirpath = os.path.dirname(filepath)
        if dirpath != "":
            os.makedirs(dirpath, exist_ok=True)

        # open the connection
        self.conn: sqlite3.Connection = sqlite3.connect(filepath, timeout=0, isolation_level=None)

    def acquire(self, timeout: Optional[float] = None) -> None:
        """Acquire the lock.

        Args:
            timeout: Optional timeout in seconds. If None, uses the default timeout set during initialization.

        Raises:
            TimeoutError: If the lock could not be acquired within the specified timeout.
        """

        used_timeout = self.timeout
        if timeout is not None:
            used_timeout = timeout
        start_time = time.time()

        while True:
            try:
                self.conn.execute("BEGIN EXCLUSIVE")
                self._is_locked = True
                # Success: we now hold the lock until we end the transaction.
                break
            except sqlite3.OperationalError:
                if (time.time() - start_time) >= used_timeout:
                    raise TimeoutError(
                        f"Could not acquire SqliteFileLock within {used_timeout} seconds."
                    )
                time.sleep(self.poll_interval)

    def release(self) -> None:
        """Release the lock."""

        try:
            self._is_locked = False
            self.conn.execute("ROLLBACK")
        except (sqlite3.OperationalError, sqlite3.ProgrammingError):
            pass

    def is_locked(self) -> bool:
        """Check if the lock is currently held by any process."""

        if self._is_locked:
            return True

        is_locked: bool = False
        try:
            self.acquire(timeout=0)
            is_locked = False
            self.release()
        except TimeoutError:
            is_locked = True
        return is_locked

    def __enter__(self) -> None:
        self.acquire()

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        self.release()

    def __del__(self) -> None:
        try:
            self.release()
            self.conn.close()
        except (sqlite3.OperationalError, sqlite3.ProgrammingError, AttributeError):
            pass



"""Functions to store, load and apply a column averaging kernel."""

from __future__ import annotations
from typing import Any, Optional
import numpy as np
import tum_esm_utils.files


class ColumnAveragingKernel:
    """A class to store, load and apply a column averaging kernel."""

    def __init__(
        self,
        szas: np.ndarray[Any, Any],
        pressures: np.ndarray[Any, Any],
        aks: Optional[np.ndarray[Any, Any]] = None,
    ) -> None:
        """Initialize the ColumnAveragingKernel.
        Args:
            szas: The solar zenith angles (SZAs) in degrees.
            pressures: The pressures in hPa.
            aks: The averaging kernels. If None, a zero array is created.
        """

        import scipy.interpolate

        self.szas = szas
        self.pressures = pressures
        self.aks: np.ndarray[Any, Any]
        if aks is not None:
            self.aks = aks
        else:
            self.aks = np.zeros((len(szas), len(pressures)), dtype=np.float64)
        self.spline: Optional[scipy.interpolate.RectBivariateSpline] = None

    def apply(
        self,
        szas: np.ndarray[Any, Any],
        pressures: np.ndarray[Any, Any],
    ) -> np.ndarray[Any, Any]:
        """Compute the averaging kernel for a given set of szas and pressures.

        ```python
        ak.apply(
            szas=np.array([0, 10, 20]),
            pressures=np.array([900, 800, 700])
        )
        ```

        Returns:

        ```
        [
           AK @  0° SZA and 900 hPa,
           AK @ 10° SZA and 800 hPa,
           AK @ 20° SZA and 700 hPa
        ]
        ```
        """

        import scipy.interpolate

        if self.spline is None:
            self.spline = scipy.interpolate.RectBivariateSpline(
                self.szas,
                self.pressures,
                self.aks,
                s=0,
            )
        return self.spline(szas, pressures, grid=False)

    def dump(self, filepath: str) -> None:
        """Dump the ColumnAveragingKernel to a JSON file."""

        assert filepath.endswith(".json"), "Filepath must end with .json"
        tum_esm_utils.files.dump_json_file(
            filepath,
            {
                "pressures": self.pressures.tolist(),
                "szas": self.szas.tolist(),
                "aks": self.aks.tolist(),
            },
            indent=None,
        )

    @staticmethod
    def load(filepath: str) -> ColumnAveragingKernel:
        """Load the ColumnAveragingKernel from a JSON file."""

        assert filepath.endswith(".json"), "Filepath must end with .json"
        d = tum_esm_utils.files.load_json_file(filepath)
        cak = ColumnAveragingKernel(
            pressures=np.array(d["pressures"], dtype=np.float64),
            szas=np.array(d["szas"], dtype=np.float64),
            aks=np.array(d["aks"], dtype=np.float64),
        )
        cak.apply(np.array([]), np.array([]))
        return cak
